/*
  File:             xxx.cpp
  Author:           Jiri Madera and Jaroslav kruis
  Purpose:          c
  Source:           modified Kunzel model
  Assumptions:
  1.9.2014
  
  
 */



#include "moistheat.h"
#include "globalt.h"
#include "math.h"
#include "globmatt.h"

dampermeability moistheatmat::damper;

moistheatmat::moistheatmat()
{
  long i;
  
  //  density of solid material
  rho_m = 0.0;
  //  water density
  rho_w = 1000.0;
  //  water molecular weight
  M = 0.01801528;
  // universal gas constant
  R = 8.314472;
  
  // permeability influenced by damage - default option is off
  daminfl = off;
}

moistheatmat::~moistheatmat()
{
}

/**
   function reads material characteristics
   
   @param in - input file

   JM, 12.10.2011
*/
void moistheatmat::read (XFILE *in)
{
  //  density
  rho.read (in);

  //  permeability
  dpp.read (in);

  //  transport coefficient
  transpar.read (in);

  //  sorption isotherm - akumulation curfe w(pv)
  sorpiso.read (in);

  //  saturated moisture
  sm.read (in);

  //  specific heat capacity
  c.read (in);

  //  thermal conductivity
  lambda.read (in);

  // temper of transport parametr
  teptr.read(in);

  xfscanf (in,"%m", &answertype_kwdset, &daminfl);
}


/**
   function prints material characteristics
   
   @param out - output file

   JM, 12.10.2011
*/
void moistheatmat::print (FILE *out)
{
  //  density
  rho.print (out);

  //  permeability
  dpp.print (out);

  //  transport coefficient
  transpar.print (out);

  //  sorption isotherm - akumulation curfe w(pv)
  sorpiso.print (out);

  //  saturated moisture
  sm.print (out);

  //  specific heat capacity
  c.print (out);

  //  thermal conductivity
  lambda.print (out);

  // temper of transport parametr
  teptr.print(out);

  fprintf(out, "%d\n", daminfl);
}

/**
   function computes conductivity %matrix of the material
   in the required integration point
   
   @param d - conductivity %matrix of material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point

*/
void moistheatmat::matcond (matrix &d,long ri,long ci,long ipp)
{
  long n;
  n = d.n;

  switch (n){
  case 1:{
    matcond1d (d,ri,ci,ipp);//1D
    break;
  }
  case 2:{
    matcond2d (d,ri,ci,ipp);//2D
    break;
  }
  case 3:{
    matcond3d (d,ri,ci,ipp);//3D
    break;
  }
  default:{
    print_err("unknown number of components of conductivity tensor is required",__FILE__,__LINE__,__func__);
  }
  }
  
  if (daminfl == on){
    if (ri==0 && ci==0){
      //  conductivity matrix is modified due to damage
      damper.matcond (d,ipp);
      Tm->ip[ipp].eqother[3] = d[0][0];
    }
    if (ri==1 && ci==0){
      //  conductivity matrix is modified due to damage
      damper.matcond (d,ipp);
      Tm->ip[ipp].eqother[4] = d[0][0];
    }
    
  }
}


/**
   function creates conductivity %matrix of the material for 1D problems

   @param d   - conductivity %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point

    JM, 12.10.2011
*/
void moistheatmat::matcond1d (matrix &d,long ri,long ci,long ipp)
{
  double k;
  k = 0.0;

  if((ri == 0) && (ci == 0))
    k = kmm(ipp);
  if((ri == 0) && (ci == 1))
    k = kmt(ipp);
  if((ri == 1) && (ci == 0))
    k = khm(ipp);
  if((ri == 1) && (ci == 1))
    k = kht(ipp);
  
  d[0][0] = k;
}

/**
   function creates conductivity %matrix of the material for 2D problems

   @param d   - conductivity %matrix of the material
   @param ri  - row index
   @param ci  - column index
   @param ipp - number of integration point

    JM, 12.10.2011
*/
void moistheatmat::matcond2d (matrix &d,long ri,long ci,long ipp)
{
  double k;
  k = 0.0;


  if((ri == 0) && (ci == 0))
    k = kmm(ipp);
  if((ri == 0) && (ci == 1))
    k = kmt(ipp);
  if((ri == 1) && (ci == 0))
    k = khm(ipp);
  if((ri == 1) && (ci == 1))
    k = kht(ipp);
  
  fillm(0.0,d);

 /* if (ipp==2)
    fprintf(Outt,"\n ri %ld ci %ld  k  %20.15le\n",ri,ci,k);
  */
  d[0][0] = k;   d[0][1] = 0.0;
  d[1][0] = 0.0; d[1][1] = k;
}

/**
   function creates conductivity %matrix of the material for 3D problems

   @param d   - conductivity %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
   
   JM, 12.10.2011
*/

void moistheatmat::matcond3d (matrix &d,long ri,long ci,long ipp)
{
  double k;
  k = 0.0;


  if((ri == 0) && (ci == 0))
    k = kmm(ipp);
  if((ri == 0) && (ci == 1))
    k = kmt(ipp);
  if((ri == 1) && (ci == 0))
    k = khm(ipp);
  if((ri == 1) && (ci == 1))
    k = kht(ipp);
  
  fillm(0.0,d);
  
  d[0][0]=k;   d[0][1]=0.0; d[0][2]=0.0;
  d[1][0]=0.0; d[1][1]=k;   d[1][2]=0.0;
  d[2][0]=0.0; d[2][1]=0.0; d[2][2]=k;
}



/**
   function creates capacity %matrix of the material

   @param c   - capacity %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
    JM, 12.10.2011
*/
void moistheatmat::matcap (double &c,long ri,long ci,long ipp)
{
 c=0.0;
  
  if((ri == 0) && (ci == 0))
    c =  cmm(ipp);
  if((ri == 0) && (ci == 1))
    c = cmt(ipp);
  if((ri == 1) && (ci == 0))
    c = chm(ipp);
  if((ri == 1) && (ci == 1)){
    c = cht(ipp);
  }
}

/**
   function checks values of relative humidity and temperature
   if the values are out of range, the limit values are assigned
   
   @param nv - array with relative humidity and temperature
   
   JK, 7.10.2011
*/
void moistheatmat::values_correction (vector &nv)
{
  double rh, tk, pv, pvs;
  
	// partial pressure of water vapor
	pv=nv[0];
	//  temperature
	tk=nv[1];
// tady to udelat lepe....
	if(pv >= 3.16995416e+03)

	  pv = 3.1699e+03;
	if(pv <= 3.17014677e-06)
	  pv = 3.18e-06;

	nv[0]=pv;
	nv[1]=tk;

}

//---------------------------------------------------------------------------
/**
   conductivity coefficient between moisture gradient and moisture flux
   
   @param ipp - id of integration point

   JM, 12.10.2011
*/
double moistheatmat::kmm (long ipp)
{
  double kk, transp, pv;
  
	// partial pressure of water vapor
	pv = Tm->ip[ipp].av[0];

	transp = transpar.getval (pv);

	kk = transp;
	
return (kk);
}

//---------------------------------------------------------------------------

/**
   conductivity coefficient between temperature gradient and moisture flux


   @param ipp - id of integration point

	JM, 12.10.2011
*/
double moistheatmat::kmt(long ipp)
{
  double kk;
  
     kk = 0.0;
  return (kk);
}

//---------------------------------------------------------------------------
/**
   conductivity coefficient between temperature gradient and heat flux

   @param ipp - id of integration point

    JM, 12.10.2011

 */
double moistheatmat::kht(long ipp)
{
  double moist, lamb, kk;
	// volumetric moisture content
	moist =  Tm->ip[ipp].eqother[0];
  
	lamb = lambda.getval (moist);

	//lambda+ L_v delta_p R Rov/M
	kk = lamb;

  return (kk);
}

//---------------------------------------------------------------------------

/**
   conductivity coefficient between moisture gradient and heat flux
   
   @param ipp - id of integration point
   
   JM, 12.10.2011
*/
double moistheatmat::khm (long ipp)
{
  double kk,tk,lv,dp,pv;
  
  // temperature
	tk = Tm->ip[ipp].av[1];

	// water vapor pressure
	pv = Tm->ip[ipp].av[0];

	//CorD(10,kd,moist,lambda,a2,a3);
	dp = dpp.getval (pv);

	//dp = water_vapour_permeability(tk,ipp);
	lv = latent_heat_of_evaporation_of_water (tk);

	kk = lv*dp;
  	
  return (kk);
}

//---------------------------------------------------------------------------
/**
   capacity coefficient between moisture gradient and moisture flux


   @param ipp - id of integration point

    JM, 12.10.2011
*/
double moistheatmat::cmm(long ipp)
{
  double cc,tk;
  double pv, t;
	// partial pressure of water vapor
	 pv = Tm->ip[ipp].av[0];
	// temperature
	// tk = Tm->ip[ipp].av[1];

	// akumulacni parametr - derivative
	//akumpar = akump.getval (moist);
	cc = Tm->ip[ipp].eqother[1]*rho_w;
 /*	t = Tp->time;
	if (t>1.1e5) {
	   fprintf(Outt,"cc je %3.2e pro pv= %lf\n",cc,pv);
	}
   */
	//cc = akumpar;
	// testovani
	//  cc = 1.0;

	return (cc);
}

//---------------------------------------------------------------------------
/**
   capacity coefficient between temperature gradient and moisture flux


   @param ipp - id of integration point

    JM, 12.10.2011
*/
double moistheatmat::cmt(long ipp)
{
  double cc;

  cc = 0.0;
	return (cc);
}

//---------------------------------------------------------------------------
/**
   capacity coefficient between moisture gradient and heat flux


   @param ipp - id of integration point

    JM, 12.10.2011
*/
double moistheatmat::chm(long ipp)
{
  double cc;
  
      cc = 0.0;
  
  return (cc);
}

 //---------------------------------------------------------------------------
/**
   capacity coefficient between temperature gradient and heat flux


   @param ipp - id of integration point

    JM, 12.10.2011
*/
double moistheatmat::cht(long ipp)
{
  double cc;

    cc = derivative_of_the_enthalpy_density(ipp);
    
  return (cc);
}

//---------------------------------------------------------------------------
/*
   function computes the latent heat of evaporation
   
   @param tk - temperature
*/
double moistheatmat::latent_heat_of_evaporation_of_water(double tk)
{
  return (2.5008e6)*pow((273.15/tk),(0.167+tk*3.67e-4));
}


//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
/**
   Function calculates saturation water vapor pressure
   saturation water vapor pressure by xxxx
   tk ... temperature [K]

   
   @param tk - temperature [K]
*/
double  moistheatmat::saturated_water_vapor_pressure(double tk)
{
     return (exp(23.5771 - 4042.9/(tk - 37.58)));
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

/**
   Function calculates derivative of the enthalpy density
   
   @param tk - temperature [K]
   @param ipp - id of integration point
   
   JM, 12.10.2011
*/
double moistheatmat::derivative_of_the_enthalpy_density (long ipp)
{
  double cap, moist,rho_m;

  // volumetric moisture content
  moist =  Tm->ip[ipp].eqother[0];
  
  // cap - specific heat capacity
  cap = c.getval (moist);
  //  density
  rho_m = rho.getval (0.0);
  
  return (rho_m * cap);
}



//---------------------------------------------------------------------------
//---------------------------------------------------------------------------



/**
   function computes new nodal value (for transmission_vector)
   for boundary condition (third kind of boundary condition)
   
   @param nodval - prescribed nodal value
   @param ri - row index
   @param ci - column index
   @param nid - node id
   @param bc - type of the boundary condition
   
*/
double moistheatmat::transmission_nodval (double nodval,long ri,long ci,long nid,long bc)
{

  long k;
  double pv,new_nodval,h,t;
  new_nodval = 0.0;
  
  //partial pressure of water vapor
  pv = nodalval (0,nid,0);
  //  temperature
  t = nodalval (0,nid,1);

  if((ri == 0) && (ci == 0))
	new_nodval = get_transmission_nodval_hh (nodval,pv,t,bc);
  if((ri == 0) && (ci == 1))
	new_nodval = 0.0;

  if((ri == 1) && (ci == 0))
	new_nodval = get_transmission_nodval_th (nodval,bc);
  if((ri == 1) && (ci == 1))
	new_nodval = get_transmission_nodval_tt (nodval,bc);

  return (new_nodval);
}

/**
   function computes the right prescribed value on the boundary for prescribed condition (partial pressure of water vapor)

   @param bv - prescribed value near the boundary
   @param rh - relative humidity in domain solved
   @param t - temperature in domain solved
   @param bc - type of boundary condition
   
*/
double moistheatmat::get_transmission_nodval_hh (double bv,double rh,double t,long bc)
{
  double nodval,pgws,moist;

  switch (bc){
  case 5:{
    nodval = bv;
    break;
  }
  case 30:{
	// relative humidity -> pgw

   //	pgws = saturated_water_vapor_pressure(t);
   //	nodval = bv*pgws;
	nodval = bv;
	break;
  }
  case 31:{
	//pgw
	nodval = bv;
	break;
  }
  case 32:{
	//relative humidity -> pgw

	pgws = saturated_water_vapor_pressure(t);
	nodval = pgws*rh;
	bv = pgws*bv;
	//inverse eq.
	nodval = bv - nodval;

	break;
  }
  case 33:{
	//h -measurements

	//  density
	rho_m = rho.getval (0.0);
	// volumetric moisture content
	moist = bv/100.0*rho_m/rho_w;
	//relative humidity from inverse sorption isotherm;
	//inverse_sorption_izoterms_giva_data(0,nodval,t,moist,a3,ipp);
	nodval = sorpiso.inverse_isotherm_value (moist);
	break;
  }
  default:{
    print_err("no acceptable boundary condition is prescribed",__FILE__,__LINE__,__func__);
    exit(0);
  }
  }
  return nodval;
}

/**
   function computes the right prescribed value on the boundary for prescribed condition (relative humidity)

   @param bv - prescribed value near the boundary
   @param rh - relative humidity in domain solved
   @param t - temperature in domain solved
   @param bc - type of boundary condition
   
*/
double moistheatmat::get_transmission_nodval_th (double bv,long bc)
{
  double nodval,pgws,moist;

  switch (bc){
  case 5:{
    nodval = bv;
    break;
  }
  default:{
    print_err("no acceptable boundary condition is prescribed",__FILE__,__LINE__,__func__);
    exit(0);
  }
  }
  return nodval;
}

/**
   function returns prescribed temperature of external environment for the transmisson boundary condition

   @param bv - prescribed external temperature
   @param bc - type of boundary condition
   
   TKr, modified by JK
*/
double moistheatmat::get_transmission_nodval_tt (double bv,long bc)
{
  double nodval;
  
  switch (bc){
  case 5:{
    nodval = bv;
    break;
  }
  case 30:{
    nodval = bv;
    break;
  }
  default:{
    print_err("no acceptable boundary condition is prescribed",__FILE__,__LINE__,__func__);
    exit(0);
  }
  }
  
  return (nodval);
}



/**
   function computes new transmission coefficient
   
   @param trc - prescribed transmission coefrhcient on the boundary
   @param ri  - row index
   @param ci  - column index
   @param nid - number of node
   @param bc  - type of boundary condition
   
*/
double moistheatmat::transmission_transcoeff (double trc,long ri,long ci,long nid,long bc)
{
  //long k;
  double new_trc;
  //double pv,t;
  new_trc = 0.0;
  /*
  // partial pressure of water vapor
  pv = nodalval (0,nid,0);
  //  temperature
  t = nodalval (0,nid,1);
  */
  if((ri == 0) && (ci == 0))
	new_trc = get_transmission_transcoeff_hh (trc,bc);
  if((ri == 0) && (ci == 1))
	new_trc = 0.0;

  if((ri == 1) && (ci == 0))
	new_trc = get_transmission_transcoeff_th (trc,bc);
  if((ri == 1) && (ci == 1))
	new_trc = get_transmission_transcoeff_tt (trc,bc);

  return (new_trc);
}

/**
   function creates transfer coefrhcient on the boundary for prescribed condition (relative humidity)

   @param t - temperature
   @param bc - type of boundary condition

*/
double moistheatmat::get_transmission_transcoeff_hh (double t,long bc)
{
  double trc,pgws;

  switch (bc){
  case 5:{
    trc = t;
    break;
  }
  case 30:{
	// relative humidity
	//pgws = saturated_water_vapor_pressure (t);
	//trc = pgws;
	trc = t;
	break;
  }
  case 31:{
	//water vapour pressure pgw
	pgws = saturated_water_vapor_pressure (t);
	trc = pgws;
	trc = 1.0;
	break;
  }
  case 32:{
	//water vapour pressure pgw
	trc = 0.0;
	break;
  }
  case 33:{
	//h - measurements
	trc = 1.0;
	break;
  }
  default:{
	print_err("no acceptable boundary condition is prescribed",__FILE__,__LINE__,__func__);
	exit(0);
  }
  }

  return trc;
}

/**
   function creates transfer coefficient on the boundary for prescribed condition (relative humidity)

   @param t - temperature
   @param bc - type of boundary condition

*/
double moistheatmat::get_transmission_transcoeff_th (double t,long bc)
{
  double trc;

  switch (bc){
  case 5:{
    trc = t;
    break;
  }
  default:{
    print_err("no acceptable boundary condition is prescribed",__FILE__,__LINE__,__func__);
    exit(0);
  }
  }
  
  return trc;
}

/**
   function creates transfer coefrhcient on the boundary for prescribed condition (temperature)

   @param trc - prescribed transmission coefficent
   @param bc  - type of boundary condition

*/
double moistheatmat::get_transmission_transcoeff_tt (double trcp,long bc)
{
  double trc;

  switch (bc){
  case 5:{
    trc = trcp;
  }
  case 30:{
	//heat transmission
	trc = trcp;
	break;
  }
  default:{
	print_err("no acceptable boundary condition is prescribed",__FILE__,__LINE__,__func__);
	exit(0);
  }
  }

  return trc;
}



/**
   function computes flux (for transmission_vector)
   for boundary condition (third kind of boundary condition)

   @param nodval - prescribed nodal value
   @param ri - row index
   @param ci - column index
   @param nid - node id
   @param bc - type of boundary condition

*/
double moistheatmat::transmission_flux (double nodval,long ri,long ci,long nid,long bc)
{
  long k;
  double flux,pv,t;

  flux = 0.0;

  // partial pressure of water vapor
  pv = nodalval (0,nid,0);
  //  temperature
  t = nodalval (0,nid,1);

  if((ri == 0) && (ci == 0))
	flux = get_transmission_flux_hh (nodval,pv,t,bc);
  if((ri == 0) && (ci == 1))
	flux = 0.0;

  if((ri == 1) && (ci == 0))
	flux = 0.0;
  if((ri == 1) && (ci == 1))
	flux = get_transmission_flux_tt (nodval,t,bc);

  return flux;
}

/**
   function creates flux on the boundary (convective mass transfer)

   @param bv - prescribed value near the boundary
   @param pv - relative humidity in domain solved
   @param t - temperature in domain solved
   @param bc - type of boundary condition

*/
double moistheatmat::get_transmission_flux_hh (double bv,double pv,double t,long bc)
{
  /* tady to udelat znova
  double flux,pgws,moist,nodval;

  switch (bc){//type of prescribed variable
  case 30:{
	//relative humidity -> pgw
	pgws = saturated_water_vapor_pressure (t);

	flux = pgws*rh;
	bv = pgws*bv;
	flux = bv - flux;//inverse eq.

	break;
  }
  case 31:{
	//water vapour pressure pgw
	pgws = saturated_water_vapor_pressure (t);
	flux = pgws*rh;
	flux = bv - flux;
	break;
  }
  case 32:{
	//relative humidity -> pgw
	pgws = saturated_water_vapor_pressure(t);
	flux = pgws*rh;
	bv = pgws*bv;
	flux = bv - flux;//inverse eq.

	break;
  }
  case 33:{
	//h -measurements

	//  density
	rho_m = rho.getval (0.0);
	// volumetric moisture content
	moist = bv/100*rho_m/rho_w;

	//relative humidity from inverse sorption isotherm;
	//inverse_sorption_izoterms_giva_data(0,nodval,t,moist,a3,ipp);
	nodval = sorpiso.inverse_isotherm_value (moist);

	flux = nodval - rh;
	break;
  }
  default:{
	print_err("no acceptable boundary condition is prescribed",__FILE__,__LINE__,__func__);
	exit(0);
  }
  }

  return flux;  */
}




/**
   function creates heat flux on the boundary

   @param bv - prescribed value near the boundary
   @param t - actual temperature on the boundary
   @param bc - type of boundary condition

*/
double moistheatmat::get_transmission_flux_tt (double bv,double t,long bc)
{
  double flux;

  switch (bc){//type of prescribed variable
  case 30:{
	//heat transmission
	flux = (bv - t);
	break;
  }
  default:{
	print_err("no acceptable boundary condition is prescribed",__FILE__,__LINE__,__func__);
	exit(0);
  }
  }
  return(flux);
}


/**
   function computes all variables in nodes
   @param compother - number of other components
   @param ipp - rhrst integration point on element
   @param pc - capillary pressure on actual node
   @param pg - gas pressure on actual node
   @param t - temperature on actual node
*/

double moistheatmat::get_othervalue(long compother,double rh,double t, long ipp)
{
  double other,w;

  switch (compother){
  case 0:{//relative humidity
    other = rh/saturated_water_vapor_pressure (teptr.getval(0));
	break;
  }
  case 1:{//temperature
	other = t-273.15;
	  break;
  }
  case 2:{//moisture content w = m3/m3
	//sorption_izoterms_giva_data(0,rh,t,w,a3, ipp);
	w = sorpiso.isotherm_value (rh);
	other = w;
	break;
  }
  case 3:{//water vapour pressure pgw
	other = rh * saturated_water_vapor_pressure (t);
	break;
  }
  case 4:{//moisture content u = kg/kg (u.rho_m/rho_w = w)
	//  sorption_izoterms_giva_data(0,rh,t,w,a3, ipp);
	rho_m = rho.getval (0.0);
	other = w*rho_w/rho_m;
	break;
  }
  case 5:{
  	  t = 298.15;
	  other = rh/saturated_water_vapor_pressure (t);
	  break;
  }
  case 6:{//temperature
	  other = rh*saturated_water_vapor_pressure (t)*M/(R*t);
	  break;
  }
  case 7:{//temperature
	  other = rh*R*t/(saturated_water_vapor_pressure (t)*M);
	  break;
  }
  case 8:{//temperature
	  other = 0.0;
	  break;
  }
  case 9:{//temperature
	  other = 0.0;
	  break;
  }
  case 10:{//temperature
	  other = 0.0;
	  break;
  }
  default:{
	print_err("unknown type of component is required",__FILE__,__LINE__,__func__);
  }
  }
  return (other);

}

/**
     function prints names of all variables in nodes
     @param out - output rhle
     @param compother - number of other components
*/
void moistheatmat::print_othervalue_name(FILE *out,long compother)
{
  switch (compother){
  case 0:{//Relative humidity
    fprintf (out,"Relative humidity ()       ");
    break;
  }
  case 1:{//Temperature
    fprintf (out,"Temperature (K)             ");
	break;
  }
  case 2:{//Moisture content w
   fprintf (out,"Moisture content w (m3/m3)   ");
    break;
  }
  case 3:{//water vapour pressure pgw
    fprintf (out,"Water vapour pressure (Pa)  ");
    break;
  }
  case 4:{//Moisture content u 
   fprintf (out,"Moisture content u (kg/kg)   ");
    break;
  }
  default:{
    print_err("unknown type of component is required",__FILE__,__LINE__,__func__);
  }
  }
}


//---------------------------------------------------------------------------


//---------------------------------------------------------------------------

/**
   function computes auxilary values
   all input values have to be imported as parameters of the function
   because it is used in connection with integration points and nodes
   therefore, the input values cannot be read in this function
   
   @param ipp - integration point id

   JM, 12.10.2011
*/
void moistheatmat::aux_values (long ipp,double *inv,double *inp,double *ine,double *out)
{
  double rh,tk,w,dwdf, kapaA, smc, tep;

	double pv;
	// bla bla bla
	pv = inv[0];
	// temperature
	tk = inv[1];

	//  saturated volumetric moisture content
	smc = sm.getval (0.0);
	//  w - volumetric moisture content
	w = sorpiso.isotherm_value (pv);
	//  derivative of sorption isotherm with respect to relative humidity
	dwdf = sorpiso.derivative_isotherm_value (pv);
	// temperature of transport parameter
	tep = teptr.getval(0.0);


	if (w>smc)
	  w = smc;
	if (w<0.0)
	  w = 0.0;

	// volumetric moisture content
	out[0] = w;
	// derivative of sorption isoterm
	out[1] = dwdf;
	// saturated volumetric moisture content
	out[2] = smc;
	// temperature of transport parameter
	out[3] = tep;


}

/**
   function stores values to eqother arrays in integration points

   @param ipp - integration point id
   @param out - array with values

   7.10.2011
*/
void moistheatmat::save_values (long ipp,double *out)
{
  Tm->ip[ipp].eqother[0]=out[0];
  Tm->ip[ipp].eqother[1]=out[1];
  Tm->ip[ipp].eqother[2]=out[2];
  
}

/**
   function reads values from arrays in integration points

   @param ipp - integration point id
   @param av - actual values
   @param eq - array with values

   7.10.2011
*/
void moistheatmat::give_values (long ipp,double *av,double *pv,double *eq)
{
  av[0] = Tm->ip[ipp].av[0];
  av[1] = Tm->ip[ipp].av[1];

  pv[0] = Tm->ip[ipp].pv[0];
  pv[1] = Tm->ip[ipp].pv[1];

  eq[0] = Tm->ip[ipp].eqother[0];
  eq[1] = Tm->ip[ipp].eqother[1];
  eq[2] = Tm->ip[ipp].eqother[2];

}

/*
  function defines initial values
  
  @param ipp - integration point id
  
  11.7.2012
*/
void moistheatmat::initvalues (long ipp,long ido)
{
}
/**
  The function returns ordered dof names of primary unknowns 
  required by the model.

  @param dofname   - array of uknown name for particular nodal dofs (output)
                     dofname[i] = name of i-th nodal unknown (for names see aliast.h - enum namevart)
  @param ntm       - number of transported media = number of nodal dof = length of array dofname

  //pv - partial pressure of water vapor
  Created by Tomas Koudelka, 4.2.2014
*/
void moistheatmat::give_dof_names(namevart *dofname, long ntm)
{
  if (ntm < 1)
  {
    print_err("the model defines %ld unknowns but number of transported media is %ld", 
              __FILE__, __LINE__, __func__, 1, ntm);
    abort();
  }
  dofname[0] = trf_press_water_vapor;
  dofname[1] = trf_temperature;
}



/**
   Function returns temperature in the given integration point.
   
   @param ipp - integration point id

   @return Funtion returns value of temperature stored in the integartion point.

   24. 10. 2013, JK
*/
double moistheatmat::give_temperature (long ipp)
{
  return Tm->ip[ipp].av[1];
}



/**  
  Function returns initial temperature in the given integration point.
 
  @param ipp - integration point id

  @return Funtion returns value of initial temperature stored in the integartion point.

  Created by TKo, 21.11.2013
*/
double moistheatmat::give_inittemperature (long ipp)
{
  print_err("Not yet implemented", __FILE__, __LINE__, __func__);
  abort();
}



/**
  Function returns relative humidity in the given integration point.

  @param ipp - integration point id

  @return Funtion returns value of relative humidity stored in the integartion point.

  Created by TKo, 21.11.2013
*/
double moistheatmat::give_rel_hum (long ipp)
{
  double pv, pvs, relhum, t;
  pv = Tm->ip[ipp].av[0];
  t =  teptr.getval(0.0);
  pvs = saturated_water_vapor_pressure (t);
  relhum = pv/pvs;

  return relhum;               // opravit
}

/**
  Function returns partial pressure of water vapor in the given integration point.

  @param ipp - integration point id

  @return Funtion returns value of partial pressure of water vapor stored in the integartion point.

  Created by TKo, 21.11.2013
*/
double moistheatmat::give_press_water_vapor (long ipp)
{
  return Tm->ip[ipp].av[0];
}



/**  
  Function returns volumetric moisture content in the given integration point.
 
  @param ipp - integration point id

  @return Funtion returns value of volumetric moisture content stored in the integartion point.

  Created by TKo, 25.11.2013
*/
double moistheatmat::give_vol_moist (long ipp)
{
  return Tm->ip[ipp].eqother[0];  // zkontrolovat
}



/**
   The funtion marks required non-transport quantities in the array antq.
   
   @param antq - array with flags for used material types
                 antq[i] = 1 => quantity type nontransquant(i+1) is required
                 antq[i] = 0 => quantity type nontransquant(i+1) is not required

   @return The function does not return anything, but it may change content of antq array.
   
   23. 4. 2014, by TKo
*/
void moistheatmat::give_reqntq(long *antq)
{
  if (daminfl == on){
	//  damage parameter
    antq[scal_iso_damage-1] = 1;
    //  process zone length
    antq[proc_zone_length-1] = 1;
    //  crack width
    antq[crack_width-1] = 1;
  }
}
